// This module makes the different routes created with @swan-io/chicane listen to the same history instance
import {
  createBrowserHistory,
  createMemoryHistory,
  createPath,
  parsePath,
} from "history";
import * as React from "react";
import { useSyncExternalStore } from "use-sync-external-store/shim";
import { areParamsArrayEqual } from "./helpers";
import { decodeLocation } from "./location";
import { Location, Search, Subscription } from "./types";

const subscriptions = new Set<Subscription>();

// From https://github.com/facebook/fbjs/blob/v2.0.0/packages/fbjs/src/core/ExecutionEnvironment.js
const canUseDOM = !!(
  typeof window !== "undefined" &&
  window.document &&
  window.document.createElement
);

export const history = canUseDOM
  ? createBrowserHistory()
  : createMemoryHistory();

let currentLocation = decodeLocation(history.location, true);
let initialLocationHasChanged = false;

if (currentLocation.toString() !== createPath(history.location)) {
  history.replace(currentLocation.toString()); // URL cleanup
}

history.listen(({ location }) => {
  const nextLocation = decodeLocation(location, false);

  // As the `encodeSearch` function guarantees a stable sorting, we can rely on a simple URL comparison
  if (nextLocation.toString() !== currentLocation.toString()) {
    initialLocationHasChanged = true;

    const searchHasChanged =
      nextLocation.raw.search !== currentLocation.raw.search;

    const search: Search = searchHasChanged ? {} : currentLocation.search;

    if (searchHasChanged) {
      for (const key in nextLocation.search) {
        if (Object.prototype.hasOwnProperty.call(nextLocation.search, key)) {
          const value = nextLocation.search[key];

          if (value == null) {
            continue;
          }

          const prevValue = currentLocation.search[key];

          if (
            prevValue == null ||
            typeof prevValue === "string" ||
            typeof value === "string" ||
            !areParamsArrayEqual(value, prevValue)
          ) {
            search[key] = value;
          } else {
            // Reuse array instance if the new content is similar
            search[key] = prevValue;
          }
        }
      }
    }

    // Create a new location object instance
    currentLocation = {
      key: nextLocation.key,

      path:
        nextLocation.raw.path !== currentLocation.raw.path
          ? nextLocation.path
          : currentLocation.path,
      search,
      ...(nextLocation.hash != null && {
        hash: nextLocation.hash,
      }),

      raw: nextLocation.raw,
      toString: nextLocation.toString,
    };

    subscriptions.forEach((subscription) => subscription(currentLocation));
  }
});

export const subscribeToLocation = (
  subscription: Subscription,
): (() => void) => {
  subscriptions.add(subscription);

  return () => {
    subscriptions.delete(subscription);
  };
};

export const getLocation = (): Location => currentLocation;
export const hasInitialLocationChanged = () => initialLocationHasChanged;

const GetUniversalLocationContext =
  React.createContext<() => Location>(getLocation);

export const GetUniversalLocationProvider =
  GetUniversalLocationContext.Provider;

export const useGetUniversalLocation = () =>
  React.useContext(GetUniversalLocationContext);

export const useLocation = (): Location => {
  const getUniversalLocation = useGetUniversalLocation();

  return useSyncExternalStore(
    subscribeToLocation,
    getUniversalLocation,
    getUniversalLocation,
  );
};

export const pushUnsafe = (url: string): void => {
  const { pathname = "", search = "", hash = "" } = parsePath(url);
  history.push({ pathname, search, hash });
};

export const replaceUnsafe = (url: string) => {
  const { pathname = "", search = "", hash = "" } = parsePath(url);
  history.replace({ pathname, search, hash });
};

// For testing purposes
export const resetInitialHasLocationChanged = () => {
  initialLocationHasChanged = false;
};
